/*
 * TP SAT
 * PK 2023-07 for the basic functions
 * Design limitation: only supports < 128 variables (this is not really a problem, since the runtime for more than that-many variables would be unreasonable anyways)
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <stdint.h>
#include <stdbool.h>
#include <immintrin.h>

#include "xoshiro.h"

/*
 * Generic slice operations
 * Two suggested slice sizes: 64 bits (using uint64_t integers) and 256 bits (using __m256i vector types (requires AVX(2)))
 */

#ifndef SL256
typedef uint64_t slice;
const int sl_sz = 64;
const int sl_lsz = 6;
#else
typedef __m256i slice;
const int sl_sz = 256;
const int sl_lsz = 8;
#endif

uint64_t rand64(void)
{
    return xoshiro256starstar_random();
}

slice not(slice x);

slice and(slice x, slice y);

// ~x & y
slice andnot(slice x, slice y);

slice xor(slice x, slice y);

slice or(slice x, slice y);

// return @x with a 1 (resp. 0) written
// at position @bit (indexed from 0 to sl_sz - 1)
// if val is non-zero (resp. zero)
slice write_bit(slice x, int bit, int val);

// return the slice with all bits equal to zero
slice zero(void);

// return the slice with all bits equal to one
slice ones(void);

// the number of non-zero bits in @x
uint64_t popcount(slice x);

bool equal(slice x, slice y);

// true iff. all the bits of the slice are equal to zero
bool is_zero(slice x);

// @s indicates which assignements in the bitsliced @vbs are valid
// pick a valid solution from @vbs and copy it to @sol in a non-bitsliced format
void pick_sol(int8_t nvar, slice vbs[nvar], slice s, int8_t sol[nvar]);

void print_slice(slice s);

#ifndef SL256
slice not(slice x)
{
}

slice and(slice x, slice y)
{
}

slice andnot(slice x, slice y)
{
}

slice xor(slice x, slice y)
{
}

slice or(slice x, slice y)
{
}

slice write_bit(slice x, int bit, int val)
{
}

slice zero(void)
{
}

slice ones(void)
{
}

uint64_t popcount(slice x)
{
}

bool equal(slice x, slice y)
{
}

bool is_zero(slice x)
{
}

void pick_sol(int8_t nvar, slice vbs[nvar], slice s, int8_t sol[nvar])
{
    assert(s != 0);
    for (int b = 0; b < sl_sz; b++)
    {
        if ((s >> b) & 1) // slice for bit b is a valid solution
        {
            for (int v = 0; v < nvar; v++)
            {
                sol[v] = (vbs[v] >> b) & 1;
            }
            return;
        }
    }
}

void print_slice(slice s)
{
    printf("%016llX\n", s);
}
#else
const __m256i uns = {~0ULL, ~0ULL, ~0ULL, ~0ULL};

slice not(slice x)
{
}

slice and(slice x, slice y)
{
}

slice andnot(slice x, slice y)
{
}

slice xor(slice x, slice y)
{
}

slice or(slice x, slice y)
{
}

slice write_bit(slice x, int bit, int val)
{
}

slice zero(void)
{
}

slice ones(void)
{
    return uns;
}

uint64_t popcount(slice x)
{
}

bool equal(slice x, slice y)
{
}

bool is_zero(slice x)
{
}

void pick_sol(int8_t nvar, slice vbs[nvar], slice s, int8_t sol[nvar])
{
}

void print_slice(slice s)
{
    uint64_t tt[4];
    _mm256_storeu_si256(tt, s);

    printf("%016llX%016llX%016llX%016llX\n", tt[3], tt[2], tt[1], tt[0]);
}
#endif



/*
 * Represents the (OR) clause (of size 3) z(a) \/ z(b) \/ z(c),
 * in the variables z_i (0 < i < 64), where z(i) maps to z_i if i > 0, and ~z_i if i < 0
 */
typedef struct clause
{
    int8_t a;
    int8_t b;
    int8_t c;
} clause;

/*
 * NOT BITSLICED
 * Attempts to find an assignment of the @nvar (canonically named) variables
 * that satisfies the @nclause clauses represented by @cl
 * If successful, writes a satisfying assignement in @sol and returns true
 * Else returns false, and the value of @sol is undefined
 * WARNING: indexing of variables start at 1, i.e. z_i is stored at index i - 1 in sol
 */
bool _tiny_solve(int8_t nvar, size_t nclause, clause cl[nclause], int8_t sol[nvar])
{
}


/*
 * BITSLICED
 * Attempts to find an assignment of the @nvar (canonically named) variables
 * that satisfies the @nclause clauses represented by @cl
 * If successful, writes a satisfying assignement in @sol and returns true
 * Else returns false, and the value of @sol is undefined
 * if SHARP is defined, additionally count all solutions and write the result in cnt
 * WARNING: indexing of variables start at 1, i.e. z_i is stored at index i - 1 in sol
 */
bool bs_solve(int8_t nvar, size_t nclause, clause cl[nclause], int8_t sol[nvar])
{
}


/*
 * Pseudo-randomly generates a 3SAT instance in @nvar variables and @nclause clauses,
 * and writes it in @cl (which must have been allocated by the caller)
 * Clauses are guaranteed to use three distinct variables 
 * (The function is not very parcimonious in its usage of random bits, and the generated
 * instances are not quite uniform)
 */
void gen_rand_system(int8_t nvar, size_t nclause, clause cl[nclause])
{
    for (size_t c = 0; c < nclause; c++)
    {
        uint64_t r;
        int8_t   v;

        r = rand64();
        v = r & 0x800 ? (r % nvar) + 1 : - ((r % nvar) + 1);
        cl[c].a = v;

        do
        {
            r = rand64();
            v = r & 0x800 ? (r % nvar) + 1 : - ((r % nvar) + 1);
        }
        while (abs(v) == abs(cl[c].a));
        cl[c].b = v;

        do
        {
            r = rand64();
            v = r & 0x800 ? (r % nvar) + 1 : - ((r % nvar) + 1);
        }
        while ((abs(v) == abs(cl[c].a)) || (abs(v) == abs(cl[c].b)));
        cl[c].c = v;
    }
}

void print_system(int8_t nvar, size_t nclause, clause cl[nclause])
{
    printf("p cnf %d %lu\n", nvar, nclause);
    for (size_t c = 0; c < nclause; c++)
    {
        printf("%d %d %d 0\n", cl[c].a, cl[c].b, cl[c].c);
    }

    return;
}

void print_sol(int8_t nvar, size_t nclause, clause cl[nclause], int8_t sol[nvar], bool res)
{
    printf("s %s\n", res ? "SAT" : "UNSAT");

    if (res)
    {
        printf("v ");
        for (int8_t v = 0; v < nvar; v++)
            printf("%d ", sol[v] == 0 ? -(v + 1) : v + 1);
        puts("0");
    }

    return;
}
